import base64
import logging
import os
import pathlib
import tempfile
import traceback
import uuid
from dataclasses import asdict, dataclass
from os import getenv
from textwrap import dedent
from typing import List, Tuple
from urllib.parse import urljoin

import ballet.templating
import funcy as fy
import git
import requests
from ballet.exc import ConfigurationError
from ballet.project import Project
from ballet.util import truthy
from ballet.util.code import blacken_code, is_valid_python
from ballet.util.git import set_config_variables
from cookiecutter.utils import work_in
from github import Github
from notebook.notebookapp import NotebookApp
from stacklog import stacklog as _stacklog
from traitlets import Bool, Integer, Unicode, default, validate
from traitlets.config import SingletonConfigurable

TESTING_URL = 'http://some/testing/url'


@dataclass
class Response:
    result: bool
    url: str = None
    message: str = None
    tb: str = None


@dataclass
class Request:
    codeContent: str


def stacklog(level, message):
    """Stacklog decorator that uses instance method's `.logger` at given level"""
    level = logging.getLevelName(level)

    def decorator(func):
        @fy.wraps(func)
        def wrapped(self, *args, **kwargs):
            with _stacklog(fy.partial(self.log.log, level), message):
                return func(self, *args, **kwargs)
        return wrapped
    return decorator


def make_feature_and_branch_name():
    my_id = str(uuid.uuid4())
    branch_name = f'submit-feature-{my_id}'
    underscore_id = my_id.replace('-', '_')
    feature_name = underscore_id
    return feature_name, branch_name


def get_new_feature_path(changes: List[Tuple[str, str]]):
    cwd = pathlib.Path.cwd()
    for (name, kind) in changes:
        if kind == 'file' and '__init__' not in str(name):
            relname = pathlib.Path(name).relative_to(cwd)
            return relname
    return None


def make_random_state():
    return base64.urlsafe_b64encode(os.urandom(16)).decode()


@fy.decorator
def handlefailures(call):
    try:
        return call()
    except Exception as e:
        message = str(e)
        tb = ''.join(traceback.format_tb(e.__traceback__))
        return Response(result=False, message=message, tb=tb)


class AssembleApp(SingletonConfigurable):

    # -- begin traits --

    debug = Bool(
        config=True,
        help='enable debug mode (no changes made on GitHub), will read from $ASSEMBLE_DEBUG if '
             'present',
    )

    @default('debug')
    def _default_debug(self):
        _default = 'False'
        # fixme: truthy only works on strings as of ballet==0.6.11
        return truthy(getenv('ASSEMBLE_DEBUG', default=_default))

    github_token = Unicode(
        config=True,
        help='github access token, will read from $GITHUB_TOKEN if present'
    )

    @default('github_token')
    def _default_github_token(self):
        return getenv('GITHUB_TOKEN', '')

    ballet_yml_path = Unicode(
        '',
        config=True,
        help='path to ballet.yml file of Ballet project (if Lab is not run from project directory)'
    )

    @validate('ballet_yml_path')
    def _validate_ballet_yml_path(self, proposal):
        if proposal:
            return str(pathlib.Path(proposal['value']).expanduser().resolve())
        else:
            return proposal

    oauth_gateway_url = Unicode(
        'https://github-oauth-gateway.herokuapp.com',
        config=True,
        help='url to github-oauth-gateway server',
    )

    access_token_timeout = Integer(
        60,
        config=True,
        help='timeout to receive access token from server via polling'
    )

    # -- end traits --

    @fy.cached_property
    def client_id(self):
        base = self.oauth_gateway_url
        url = urljoin(base, '/api/v1/app_id')
        response = requests.get(url)
        response.raise_for_status()
        d = response.json()
        return d['client_id']

    scopes = ['read:user', 'public_repo']
    _state = None

    @property
    def state(self):
        if self._state is None:
            self._state = make_random_state()

        return self._state

    def reset_state(self):
        self._state = None

    def set_github_token(self, token):
        self.github_token = token

    _is_authenticated = False

    def is_authenticated(self):
        if not self._is_authenticated:
            with fy.suppress(Exception):
                _ = self.username
                self._is_authenticated = True

        return self._is_authenticated

    @property
    def github(self):
        return Github(self.github_token)

    @property
    def username(self):
        return self.github.get_user().login

    @property
    def useremail(self):
        # in lieu of requesting `user:email` scope
        return f'{self.username}@users.noreply.github.com'

    @property
    def reponame(self):
        """name of project repo"""
        return self.project.config.get('project.project_slug', '')

    @property
    def upstream_repo_spec(self):
        """refspec of upstream repo of the form 'username/reponame'"""
        github_owner = self.project.config.get('github.github_owner', '')
        return f'{github_owner}/{self.reponame}'

    @property
    def upstream_repo(self):
        return self.github.get_repo(self.upstream_repo_spec)

    @property
    def repo_url(self):
        """url of forked repo, including token-based authentication"""
        return f'https://{self.github_token}@github.com/{self.username}/{self.reponame}'

    @property
    def project(self):
        # 1. configuration option passed explicitly
        # 2. from notebooks dir
        # 3. from cwd
        if self.ballet_yml_path:
            return Project.from_path(self.ballet_yml_path)

        path = NotebookApp.instance().notebook_dir
        with fy.suppress(Exception):
            return Project.from_path(path)

        with fy.suppress(Exception):
            return Project.from_cwd()

        raise ConfigurationError('Could not detect Ballet project')

    @fy.post_processing(asdict)
    @handlefailures
    def create_pull_request_for_code_content(self, input_data: dict) -> Response:
        code_content = self.load_request(input_data)
        self.check_code_is_valid(code_content)

        # async
        self.fork_repo()

        with tempfile.TemporaryDirectory() as dirname:
            dirname = str(pathlib.Path(dirname).resolve())
            repo = self.clone_repo(dirname)
            with work_in(dirname):
                self.configure_repo(repo)
                feature_name, branch_name = self.create_new_branch(repo)
                changed_files, new_feature_path = self.start_new_feature(dirname, feature_name)
                self.write_code_content(new_feature_path, code_content)
                self.commit_changes(repo, changed_files)
                push_result = self.push_to_remote(repo, branch_name)  # noqa F841
                # TODO if push failed, likely because fork does not yet exist, then try again
                return self.create_pull_request(feature_name, branch_name)

    @stacklog('DEBUG', 'Loading request')
    def load_request(self, input_data: dict) -> str:
        try:
            req = Request(**input_data)
        except TypeError as e:
            raise TypeError(f'Bad request - {e}') from e
        return req.codeContent

    @stacklog('INFO', 'Checking for valid code')
    def check_code_is_valid(self, code_content: str) -> None:
        if not code_content.strip():
            raise ValueError('No code was submitted -- did you select the correct cell?')

        if not is_valid_python(code_content):
            raise ValueError('Submitted code is not valid Python code')

    @stacklog('INFO', 'Forking upstream repo')
    def fork_repo(self) -> None:
        if not self.debug:
            return self.upstream_repo.create_fork()
        else:
            self.log.debug('Didn\'t actually fork repo due to debug')
            return None

    @stacklog('INFO', 'Cloning repo')
    def clone_repo(self, dirname: str) -> git.Repo:
        return git.Repo.clone_from(self.repo_url, to_path=dirname)

    @stacklog('INFO', 'Configuring repo')
    def configure_repo(self, repo: git.Repo) -> None:
        set_config_variables(repo, {
            'user.name': self.username,  # github username
            'user.email': self.useremail,
        })
        repo.remote().set_url(self.repo_url)

    @stacklog('INFO', 'Creating new branch and checking it out')
    def create_new_branch(self, repo: git.Repo) -> Tuple[str, str]:
        feature_name, branch_name = make_feature_and_branch_name()
        repo.create_head(branch_name)
        repo.heads[branch_name].checkout()
        return feature_name, branch_name

    @stacklog('INFO', 'Starting new feature')
    def start_new_feature(self, dirname: str, feature_name: str) -> Tuple[List[str], str]:
        # start new feature
        extra_context = {
            'username': self.username.replace('-', '_'),
            'featurename': feature_name,
        }
        contrib_dir = self.project.config.get('contrib.module_path')
        changes = ballet.templating.start_new_feature(
            contrib_dir=contrib_dir,
            branching=False,
            no_input=True,
            extra_context=extra_context
        )
        changed_files = [
            str(pathlib.Path(name).relative_to(dirname))
            for (name, kind) in changes
            if kind == 'file'
        ]
        new_feature_path = get_new_feature_path(changes)
        return changed_files, new_feature_path

    @stacklog('INFO', 'Adding code content')
    def write_code_content(self, new_feature_path: str, code_content: str):
        with open(new_feature_path, 'w') as f:
            blackened_code_content = blacken_code(code_content)
            f.write(blackened_code_content)

    @stacklog('INFO', 'Committing new feature')
    def commit_changes(self, repo, changed_files):
        repo.index.add(changed_files)
        repo.index.commit('Add new feature')

    @stacklog('INFO', 'Pushing to remote')
    def push_to_remote(self, repo, branch_name) -> List[git.remote.PushInfo]:
        refspec = f'refs/heads/{branch_name}:refs/heads/{branch_name}'
        if not self.debug:
            return repo.remote().push(refspec=refspec)
        else:
            self.log.debug('Didn\'t actually push to remote due to debug')

    @stacklog('INFO', 'Creating pull request')
    def create_pull_request(self, feature_name, branch_name):
        grepo = self.github.get_repo(self.upstream_repo_spec)
        title = 'Propose new feature'
        body = dedent(f'''\
                Propose new feature: {feature_name}
                Submitted by user: {self.username}

                --
                Pull request automatically created by ballet-assemble
            ''')
        base = 'master'
        head = f'{self.username}:{branch_name}'
        maintainer_can_modify = True
        self.log.debug(
            'About to create pull: title=%s, body=%s, base=%s, head=%s',
            title, body, base, head)

        self.log.debug(
            'About to create pull: title=%s, body=%s, base=%s, head=%s',
            title, body, base, head,
        )
        if not self.debug:
            pr = grepo.create_pull(title=title, body=body, base=base, head=head,
                                   maintainer_can_modify=maintainer_can_modify)
            url = pr.html_url
        else:
            self.log.debug('Didn\'t create real pull request due to debug')
            url = TESTING_URL

        return Response(result=True, url=url)


def print_help():
    AssembleApp.class_print_help()


if __name__ == '__main__':
    print_help()



# -*- coding: utf-8 -*-

"""Server extension for ballet-assemble"""

__author__ = 'Micah Smith'
__email__ = 'micahs@mit.edu'
__version__ = '0.8.8'

from jupyterlab.labapp import LabApp

from .handlers import setup_handlers

EXTENSION_URL_PATH = 'assemble'


def _jupyter_server_extension_paths():
    return [{
        'module': 'ballet_assemble'
    }]


def load_jupyter_server_extension(app: LabApp):
    """Register the API handler

    Args:
        app (notebook.notebookapp.NotebookApp)
            Notebook application instance
    """

    # initialize app instance
    from .app import AssembleApp
    AssembleApp.clear_instance()
    AssembleApp.instance(config=app.config)

    setup_handlers(app.web_app, EXTENSION_URL_PATH)
    app.log.info('Registered ballet-assemble extension at URL path /%s',
                 EXTENSION_URL_PATH)



"""
Setup Module to setup Python Handlers for ballet-assemble.
"""
import os

from setupbase import (
    create_cmdclass, install_npm, ensure_targets,
    combine_commands, ensure_python
)
from setuptools import setup, find_packages

HERE = os.path.abspath(os.path.dirname(__file__))

# Ensure a valid python version
ensure_python('>=3.6')

# Representative files that should exist after a successful build
jstargets = [
    os.path.join(HERE, 'lib', 'serverextension.js'),
]

package_data_spec = {
    'ballet_assemble': [
        '*'
    ]
}

data_files_spec = [
    (
        'share/jupyter/lab/extensions',
        os.path.join(HERE, 'server', 'ballet_assemble', 'labextension'),
        '*.tgz',
    ),
    (
        'etc/jupyter/jupyter_notebook_config.d',
        os.path.join(HERE, 'server', 'jupyter-config'),
        'ballet-assemble.json',
    ),
]

cmdclass = create_cmdclass(
    prerelease_cmd='jsdeps',
    package_data_spec=package_data_spec,
    data_files_spec=data_files_spec
)

cmdclass['jsdeps'] = combine_commands(
    install_npm(HERE, build_cmd='build:all', npm=['jlpm']),
    ensure_targets(jstargets),
)

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

install_requires = [
    'ballet',
    'black == 19.10b0',
    'funcy',
    'gitpython',
    'jupyterlab ~= 2.0',
    'importlib-metadata; python_version < "3.8"',
    'notebook',
    'pygithub',
    'requests',
    'stacklog',
    'tenacity',
    'tornado',
]

test_requirements = [
    'coverage >= 4.5.1',
    'pytest >= 3.4.2',
    'pytest-cov >= 2.6',
]

development_requirements = [
    # general
    'bump2version >= 1.0.0',
    'pip >= 9.0.1',

    # style check
    'flake8 >= 3.5.0',
    'isort >= 4.3.4, <= 4.3.9',

    # distribute on PyPI
    'twine >= 1.10.0',
    'wheel >= 0.30.0',
]

setup_args = dict(
    name='ballet-assemble',
    version='0.8.8',
    url='https://github.com/ballet/ballet-assemble',
    author='Micah Smith',
    description='Submit ballet modules from within JupyterLab',
    long_description=long_description,
    long_description_content_type='text/markdown',
    cmdclass=cmdclass,
    packages=find_packages(where='server',
                           include=['ballet_assemble',
                                    'ballet_assemble.*']),
    package_dir={'': 'server'},
    install_requires=install_requires,
    extras_require={
        'test': test_requirements,
        'dev': development_requirements + test_requirements,
    },
    zip_safe=False,
    include_package_data=True,
    license='MIT',
    platforms='Linux, Mac OS X, Windows',
    keywords=['Jupyter', 'JupyterLab'],
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Framework :: Jupyter',
    ],
)

if __name__ == '__main__':
    setup(**setup_args)



from functools import partial
from urllib.parse import urlencode, urljoin

import requests
import tornado
from notebook.base.handlers import APIHandler, IPythonHandler
from notebook.notebookapp import NotebookWebApplication
from notebook.utils import url_path_join
from tenacity import (
    RetryError, retry, retry_if_exception_message, retry_if_exception_type, stop_after_delay,
    wait_fixed)
from tornado.httpclient import AsyncHTTPClient

from .app import AssembleApp

try:
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata


GITHUB_OAUTH_URL = 'https://github.com/login/oauth/authorize'


class StatusHandler(APIHandler):

    @tornado.web.authenticated
    def get(self):
        self.write({'status': 'OK'})


class VersionHandler(APIHandler):

    @tornado.web.authenticated
    def get(self):
        app = AssembleApp.instance()

        try:
            assemble_version = metadata.version('ballet_assemble')
        except Exception:
            assemble_version = None
        try:
            ballet_version = metadata.version('ballet')
        except Exception:
            ballet_version = None
        try:
            project_version = app.project.version
        except Exception:
            project_version = None

        self.write({
            'assemble': assemble_version,
            'ballet': ballet_version,
            'project': project_version,
        })


class ConfigHandler(APIHandler):

    @tornado.web.authenticated
    def get(self):
        app = AssembleApp.instance()
        result = {}
        for attr in app.class_own_traits():
            result[attr] = getattr(app, attr)
        self.write(result)


class ConfigItemHandler(APIHandler):

    @tornado.web.authenticated
    def get(self, attr):
        app = AssembleApp.instance()
        try:
            param = getattr(app, attr)
            self.write({attr: param})
        except AttributeError:
            self.send_error(404)


class SubmitHandler(APIHandler):

    @tornado.web.authenticated
    def post(self):
        input_data = self.get_json_body()
        app = AssembleApp.instance()
        result = app.create_pull_request_for_code_content(input_data)
        self.write(result)


class AuthorizeHandler(IPythonHandler):

    @tornado.web.authenticated
    def get(self):
        app = AssembleApp.instance()

        # wake server async
        http_client = AsyncHTTPClient()
        url = urljoin(app.oauth_gateway_url, '/status')
        http_client.fetch(url)

        # do oauth flow
        base = GITHUB_OAUTH_URL
        params = {
            'client_id': app.client_id,
            'state': app.state,
            'scope': ','.join(app.scopes),
        }

        url = base + '?' + urlencode(params)
        self.redirect(url, permanent=False)


class TokenHandler(IPythonHandler):

    @retry(
        wait=wait_fixed(3),
        retry=(
            retry_if_exception_type(RuntimeError) &
            retry_if_exception_message(match=r'[Nn]o authorization code found.*')
        ),
        stop=stop_after_delay(AssembleApp.instance().access_token_timeout)
    )
    @tornado.gen.coroutine
    def get_token(self, url, data):
        response = requests.post(url, json=data)
        d = response.json()
        if response.ok:
            # TODO also store other token info
            raise tornado.gen.Return(d['access_token'])
        else:
            reason = d.get('message', '').lower()
            raise RuntimeError(reason)

    @tornado.web.authenticated
    @tornado.gen.coroutine
    def post(self):
        """request token if we have just authenticated"""
        app = AssembleApp.instance()
        base = app.oauth_gateway_url
        state = app.state
        url = urljoin(base, '/api/v1/access_token')
        data = {'state': state}

        try:
            token = yield self.get_token(url, data)
            app.set_github_token(token)
            self.finish()
        except RetryError:
            self.send_error(status_code=400, reason='timeout')
        except RuntimeError as e:
            self.send_error(status_code=400, reason=str(e))

        app.reset_state()


class AuthenticatedHandler(APIHandler):

    @tornado.web.authenticated
    def get(self):
        app = AssembleApp.instance()
        self.write({
            'result': app.is_authenticated(),
            'message': None,
        })


def setup_handlers(app: NotebookWebApplication, url_path: str):
    host_pattern = '.*$'
    base_url = app.settings['base_url']
    route_pattern = partial(url_path_join, base_url, url_path)

    app.add_handlers(host_pattern, [
        (route_pattern('status'), StatusHandler),
        (route_pattern('version'), VersionHandler),
        (route_pattern('config'), ConfigHandler),
        (route_pattern(r'config/(.*)'), ConfigItemHandler),
        (route_pattern('submit'), SubmitHandler),
        (route_pattern('auth', 'authorize'), AuthorizeHandler),
        (route_pattern('auth', 'token'), TokenHandler),
        (route_pattern('auth', 'authenticated'), AuthenticatedHandler),
    ])
